/*
 * D3Q27AdjustForcingSimulationObserver.cpp
 *  Author: Konstantin Kutscher
 */
#include "AdjustForcingSimulationObserver.h"

#include <fstream>

#include <mpi/Communicator.h>
#include "Grid3D.h"
#include "IntegrateValuesHelper.h"
#include "UbScheduler.h"
#include <SetForcingBlockVisitor.h>

AdjustForcingSimulationObserver::AdjustForcingSimulationObserver(SPtr<Grid3D> grid, SPtr<UbScheduler> s, const std::string &path,
                                                   SPtr<IntegrateValuesHelper> integrateValues, real vTarged,
                                                   std::shared_ptr<vf::mpi::Communicator> comm)

    : SimulationObserver(grid, s), path(path), integrateValues(integrateValues), comm(comm), vx1Targed(vTarged)
{
    // cnodes = integrateValues->getCNodes();
    root = comm->isRoot();

    Ta = scheduler->getMaxStep();

    Kpcrit = 3.0 / Ta; // 0.3;
    Tcrit  = 3.0 * Ta; // 30.0;
    Tn     = 0.5 * Tcrit;
    Tv     = 0.12 * Tcrit;

    Kp = 0.6 * Kpcrit;
    Ki = Kp / Tn;
    Kd = Kp * Tv;

    y       = 0;
    e       = 0;
    esum    = 0;
    eold    = 0;
    forcing = 0;

    if (root) {
        std::string fname = path + "/forcing/forcing.csv";
        std::ofstream ostr;
        ostr.open(fname.c_str(), std::ios_base::out | std::ios_base::app);
        if (!ostr) {
            ostr.clear();
            std::string file_path = UbSystem::getPathFromString(fname);
            if (file_path.size() > 0) {
                UbSystem::makeDirectory(file_path);
                ostr.open(fname.c_str(), std::ios_base::out | std::ios_base::app);
            }
            if (!ostr)
                throw UbException(UB_EXARGS, "couldn't open file " + fname);
        }
        ostr << "step;volume;vx1average;forcing\n";
        ostr.close();

        //////////////////////////////////////////////////////////////////////////////////////////////////
        // temporary solution
        std::string fNameCfg = path + "/forcing/forcing.cfg";
        std::ifstream istr2;
        istr2.open(fNameCfg.c_str(), std::ios_base::in);
        if (istr2) {
            istr2 >> forcing;
            // istr2 >> esum;
            // istr2 >> eold;
        }
        istr2.close();
    }
    ////////////////////////////////////////////////////////////////////////////////////////////////////////
}

//////////////////////////////////////////////////////////////////////////
void AdjustForcingSimulationObserver::update(real step)
{
    if (scheduler->isDue(step))
        collectData(step);
}
//////////////////////////////////////////////////////////////////////////
void AdjustForcingSimulationObserver::collectData(real step)
{
    //////////////////////////////////////////////////////////////////////////////////////////////////
    // temporary solution
    if (root) {
        std::string fNameCfg = path + "/forcing/forcing.cfg";
        std::ofstream ostr2;
        ostr2.open(fNameCfg.c_str(), std::ios_base::out);
        if (!ostr2) {
            ostr2.clear();
            std::string path = UbSystem::getPathFromString(fNameCfg);
            if (path.size() > 0) {
                UbSystem::makeDirectory(path);
                ostr2.open(fNameCfg.c_str(), std::ios_base::out);
            }
            if (!ostr2)
                throw UbException(UB_EXARGS, "couldn't open file " + fNameCfg);
        }
        ostr2 << forcing << " " << esum << " " << eold;
        ostr2.close();
    }
    ////////////////////////////////////////////////////////////////////////////////////////////////////////

    integrateValues->calculateMQ();

    if (root) {
        cellsVolume = integrateValues->getCellsVolume();
        real vx1  = integrateValues->getVx1();
        vx1Average  = (vx1 / cellsVolume);

        //////////////////////////////////////////////////////////////////////////
        // PID-Controller (PID-Regler)
        e    = vx1Targed - vx1Average;
        esum = esum + e;
        y    = Kp * e + Ki * Ta * esum + Kd * (e - eold) / Ta;
        eold = e;

        forcing = forcing + y;
        //////////////////////////////////////////////////////////////////////////
    }
    //////////////////////////////////////////////////////////////////////////
    comm->broadcast(forcing);

    mu::Parser fctForcingX1, fctForcingX2, fctForcingX3;
    fctForcingX1.SetExpr("Fx1");
    fctForcingX1.DefineConst("Fx1", forcing);
    fctForcingX2.SetExpr("0.0");
    fctForcingX3.SetExpr("0.0");
    SetForcingBlockVisitor forcingVisitor(fctForcingX1, fctForcingX2, fctForcingX3);
    grid->accept(forcingVisitor);

    // for(CalcNodes cn : cnodes)
    //{
    //   LBMKernel3DPtr kernel = cn.block->getKernel();
    //   if (kernel)
    //   {
    //      kernel->setForcingX1(fctForcingX1);
    //      kernel->setWithForcing(true);
    //   }
    //
    //}

    if (root) {
        // UBLOG(logINFO, "D3Q27AdjustForcingSimulationObserver step: " << static_cast<int>(step));
        // UBLOG(logINFO, "new forcing is: " << forcing);
        std::string fname = path + "/forcing/forcing.csv";
        // std::string fname = path + "/forcing/forcing_"+UbSystem::toString(comm->getProcessID())+".csv";
        std::ofstream ostr;
        ostr.open(fname.c_str(), std::ios_base::out | std::ios_base::app);
        if (!ostr) {
            ostr.clear();
            std::string path = UbSystem::getPathFromString(fname);
            if (path.size() > 0) {
                UbSystem::makeDirectory(path);
                ostr.open(fname.c_str(), std::ios_base::out | std::ios_base::app);
            }
            if (!ostr)
                throw UbException(UB_EXARGS, "couldn't open file " + fname);
        }
        int istep = static_cast<int>(step);

        // ostr << istep << ";" << cellsVolume << ";" << vx1Average << "; " << forcing << "\n";
        ostr << istep << ";" << cellsVolume << ";" << vx1Average << "; " << forcing << "; " << e << "; " << esum << "; "
             << y << "\n";
        ostr.close();
    }
}
